import argparse
import contextlib
import os
import sys
import pkg_resources
from typing import Any

from iiblib import iib_client, iib_authentication


def setup_iib_client(parsed_args: argparse.Namespace) -> iib_client.IIBClient:
    iib_auth = iib_authentication.IIBKrbAuth(
        parsed_args.iib_krb_principal,
        parsed_args.iib_server,
        ktfile=parsed_args.iib_krb_ktfile,
    )
    kwargs = {
        "auth": iib_auth,
    }
    if parsed_args.iib_insecure:
        kwargs["ssl_verify"] = False
    if parsed_args.build_timeout:
        kwargs["wait_for_build_timeout"] = int(parsed_args.build_timeout)
    iibc = iib_client.IIBClient(parsed_args.iib_server, **kwargs)
    return iibc


def setup_arg_parser(args: dict[Any, Any]) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    arg_groups: dict[Any, Any] = {}
    for aliases, arg_data in args.items():
        holder = parser
        if arg_data["group"]:
            arg_groups.setdefault(
                arg_data["group"], parser.add_argument_group(arg_data["group"])
            )
            holder = arg_groups[arg_data["group"]]
        action = arg_data.get("action")
        if not action and arg_data["type"] == bool:
            action = "store_true"
        kwargs = {
            "help": arg_data.get("help"),
            "required": arg_data.get("required", False),
            "default": arg_data.get("default"),
        }
        if action:
            kwargs["action"] = action
        else:
            kwargs["type"] = arg_data.get("type", "str")
            kwargs["nargs"] = arg_data.get("count")

        holder.add_argument(*aliases, **kwargs)

    return parser


@contextlib.contextmanager
def setup_entry_point_cli(
    entry_tuple: tuple[str, str, str],
    name: str,
    args: list[str],
    environ_vars: dict[str, str],
) -> Any:
    orig_argv = sys.argv[:]
    orig_environ = os.environ.copy()

    try:
        # First argv element is always the entry point name.
        # For a console_scripts entry point, this will be the same value
        # as if the script was invoked directly. For any other kind of entry point,
        # this value is probably meaningless.
        sys.argv = [name]
        sys.argv.extend(args)
        for key in environ_vars:
            os.environ[key] = environ_vars[key]
        entry_point_func = pkg_resources.load_entry_point(*entry_tuple)
        yield entry_point_func
    finally:
        sys.argv = orig_argv[:]
        os.environ.update(orig_environ)

        to_delete = [key for key in os.environ if key not in orig_environ]
        for key in to_delete:
            del os.environ[key]



import contextlib
import logging
import mock
import pkg_resources
import pytest
import os

from pubtools.iib.utils import setup_entry_point_cli
from pubtools.iib.iib_ops import _iib_op_main, print_error_message

import pushcollector
import requests_mock


from iiblib.iib_build_details_model import IIBBuildDetailsModel
from more_executors.futures import f_return

from utils import FakeTaskManager, FakeCollector


FIXTURE_IIB_SERVER = "iib-server"


fake_tm = FakeTaskManager()

operator_1_push_item_pending = {
    "state": "PENDING",
    "origin": "index-image",
    "filename": "operator-1",
    "build": "feed.com/index/image:tag",
    "dest": "redhat-operator-index",
    "signing_key": None,
    "checksums": None,
    "src": "bundle1",
}
operator_1_push_item_notpushed = operator_1_push_item_pending.copy()
operator_1_push_item_notpushed["state"] = "NOTPUSHED"
operator_1_push_item_pushed = operator_1_push_item_pending.copy()
operator_1_push_item_pushed["state"] = "PUSHED"

operator_1_push_item_delete_pending = {
    "state": "PENDING",
    "origin": "index-image",
    "filename": "operator-1",
    "build": "feed.com/index/image:tag",
    "dest": "redhat-operator-index",
    "signing_key": None,
    "checksums": None,
    "src": None,
}
operator_1_push_item_deleted = operator_1_push_item_delete_pending.copy()
operator_1_push_item_deleted["state"] = "DELETED"
operator_1_push_item_delete_notpushed = operator_1_push_item_delete_pending.copy()
operator_1_push_item_delete_notpushed["state"] = "NOTPUSHED"

operator_1_push_item_ad_pending = operator_1_push_item_pending.copy()
operator_1_push_item_ad_pending["src"] = None
operator_1_push_item_ad = operator_1_push_item_ad_pending.copy()
operator_1_push_item_ad["state"] = "PUSHED"
operator_1_push_item_ad_notpushed = operator_1_push_item_ad_pending.copy()
operator_1_push_item_ad_notpushed["state"] = "NOTPUSHED"


@contextlib.contextmanager
def setup_entry_point_py(entry_tuple, environ_vars):
    orig_environ = os.environ.copy()
    try:
        for key in environ_vars:
            os.environ[key] = environ_vars[key]
        entry_point_func = pkg_resources.load_entry_point(*entry_tuple)
        yield entry_point_func
    finally:
        os.environ.update(orig_environ)


@pytest.fixture
def fixture_iib_client():
    with mock.patch("iiblib.iib_client.IIBClient") as iibc_patched:
        iibc_patched.return_value.add_bundles.side_effect = (
            lambda *args, **kwargs: IIBBuildDetailsModel.from_dict(
                fake_tm.setup_task(*args, **kwargs)
            )
        )
        iibc_patched.return_value.remove_operators.side_effect = (
            lambda *args, **kwargs: IIBBuildDetailsModel.from_dict(
                fake_tm.setup_task(*args, **dict(list(kwargs.items())+ [("op_type", "rm")]))
            )
        )
        iibc_patched.return_value.add_deprecations.side_effect = (
            lambda *args, **kwargs: IIBBuildDetailsModel.from_dict(
                fake_tm.setup_task(*args, **dict(list(kwargs.items())+ [("op_type", "add-deprecations")]))
            )
        )
        iibc_patched.return_value.get_build.side_effect = fake_tm.get_task
        iibc_patched.return_value.wait_for_build.side_effect = (
            lambda build_details: IIBBuildDetailsModel.from_dict(
                fake_tm.get_task(build_details.id)
            )
        )
        yield iibc_patched


@pytest.fixture
def fixture_pushcollector():
    fake_collector = FakeCollector()
    pushcollector.Collector.register_backend(
        "pubtools-ibb-test", lambda: fake_collector
    )
    pushcollector.Collector.set_default_backend("pubtools-ibb-test")
    yield fake_collector


@pytest.fixture
def fixture_iib_krb_auth():
    with mock.patch("iiblib.iib_authentication.IIBKrbAuth") as iib_krbauth_patched:
        iib_krbauth_patched.return_value = mock.MagicMock(name="MockedIIBKrbAuth")
        yield iib_krbauth_patched


@pytest.fixture
def fixture_common_iib_op_args():
    return [
        "--iib-server",
        FIXTURE_IIB_SERVER,
        "--index-image",
        "index-image",
        "--binary-image",
        "binary-image",
        "--iib-krb-principal",
        "example@REALM",
        "--iib-insecure",
        "--overwrite-from-index",
        "--build-tag",
        "extra-tag-1",
        "--build-tag",
        "extra-tag-2",
    ]


def add_bundles_mock_calls_tester(
    fixture_iib_client,
    fixture_iib_krb_auth,
):
    fixture_iib_client.return_value.add_bundles.assert_called_once_with(
        "index-image",
        bundles=["bundle1"],
        arches=["arch"],
        binary_image="binary-image",
        overwrite_from_index=True,
        overwrite_from_index_token="overwrite_from_index_token",
        deprecation_list=["bundle1"],
        build_tags=["extra-tag-1", "extra-tag-2"],
    )
    fixture_iib_client.assert_called_once_with(
        "iib-server",
        auth=fixture_iib_krb_auth.return_value,
        ssl_verify=False,
        wait_for_build_timeout=30,
    )


def add_bundles_mock_calls_tester_check_related_images(
    fixture_iib_client,
    fixture_iib_krb_auth,
):
    fixture_iib_client.return_value.add_bundles.assert_called_once_with(
        "index-image",
        bundles=["bundle1"],
        arches=["arch"],
        binary_image="binary-image",
        overwrite_from_index=True,
        overwrite_from_index_token="overwrite_from_index_token",
        build_tags=["extra-tag-1", "extra-tag-2"],
        check_related_images=True,
    )
    fixture_iib_client.assert_called_once_with(
        "iib-server", auth=fixture_iib_krb_auth.return_value, ssl_verify=False
    )


def add_bundles_mock_calls_tester_empty_deprecation_list(
    fixture_iib_client,
    fixture_iib_krb_auth,
):
    fixture_iib_client.return_value.add_bundles.assert_called_once_with(
        "index-image",
        bundles=["bundle1"],
        arches=["arch"],
        binary_image="binary-image",
        overwrite_from_index=True,
        overwrite_from_index_token="overwrite_from_index_token",
        build_tags=["extra-tag-1", "extra-tag-2"],
    )
    fixture_iib_client.assert_called_once_with(
        "iib-server", auth=fixture_iib_krb_auth.return_value, ssl_verify=False
    )


def add_bundles_mock_calls_tester_deprecation_bundles(
    fixture_iib_client,
    fixture_iib_krb_auth,
):
    fixture_iib_client.return_value.add_bundles.assert_called_once_with(
        "index-image",
        bundles=["bundle1"],
        arches=["arch"],
        binary_image="binary-image",
        overwrite_from_index=True,
        overwrite_from_index_token="overwrite_from_index_token",
        deprecation_list=["bundle1", "bundle2"],
        build_tags=["extra-tag-1", "extra-tag-2"],
    )
    fixture_iib_client.assert_called_once_with(
        "iib-server", auth=fixture_iib_krb_auth.return_value, ssl_verify=False
    )


def remove_operators_mock_calls_tester(
    fixture_iib_client,
    fixture_iib_krb_auth,
):
    fixture_iib_client.assert_called_once_with(
        "iib-server", auth=fixture_iib_krb_auth.return_value, ssl_verify=False
    )
    fixture_iib_client.return_value.remove_operators.assert_called_once_with(
        "index-image",
        operators=["operator-1"],
        arches=["arch"],
        binary_image="binary-image",
        overwrite_from_index=True,
        overwrite_from_index_token="overwrite_from_index_token",
        build_tags=["extra-tag-1", "extra-tag-2"],
    )


def add_deprecations_mock_calls_tester(
    fixture_iib_client,
    fixture_iib_krb_auth,
):
    fixture_iib_client.assert_called_once_with(
        "iib-server", auth=fixture_iib_krb_auth.return_value, ssl_verify=False
    )
    fixture_iib_client.return_value.add_deprecations.assert_called_once_with(
        "index-image",
        deprecation_schema='{"a": "b"}',
        operator_package="operator-1",
        binary_image="binary-image",
        overwrite_from_index=True,
        overwrite_from_index_token="overwrite_from_index_token",
        build_tags=["extra-tag-1", "extra-tag-2"],
    )


@pytest.mark.parametrize(
    "extra_args,push_items,mock_calls_tester",
    [
        (
            ["--deprecation-list", "bundle1,bundle2", "--arch", "arch"],
            [operator_1_push_item_pending, operator_1_push_item_pushed],
            add_bundles_mock_calls_tester_deprecation_bundles,
        ),
        (
            ["--deprecation-list", "bundle1", "--build-timeout", "30", "--arch", "arch"],
            [operator_1_push_item_pending, operator_1_push_item_pushed],
            add_bundles_mock_calls_tester,
        ),
        (
            ["--deprecation-list", "", "--arch", "arch"],
            [operator_1_push_item_pending, operator_1_push_item_pushed],
            add_bundles_mock_calls_tester_empty_deprecation_list,
        ),
        (
            ["--check-related-images", "--arch", "arch"],
            [operator_1_push_item_pending, operator_1_push_item_pushed],
            add_bundles_mock_calls_tester_check_related_images,
        ),
    ],
)
def test_add_bundles_cli(
    fixture_iib_client,
    fixture_iib_krb_auth,
    fixture_common_iib_op_args,
    fixture_pushcollector,
    extra_args,
    push_items,
    mock_calls_tester,
):
    with setup_entry_point_cli(
        ("pubtools_iib", "console_scripts", "pubtools-iib-add-bundles"),
        "pubtools-iib-add-bundle",
        fixture_common_iib_op_args + ["--bundle", "bundle1"] + extra_args,
        {
            "OVERWRITE_FROM_INDEX_TOKEN": "overwrite_from_index_token",
        },
    ) as entry_func:
        entry_func()
    assert fixture_pushcollector.items == push_items
    mock_calls_tester(
        fixture_iib_client,
        fixture_iib_krb_auth,
    )


@mock.patch("pubtools.iib.iib_ops.print_error_message")
def test_add_bundles_cli_error(
    mock_print_error_message,
    fixture_iib_client,
    fixture_iib_krb_auth,
    fixture_common_iib_op_args,
    fixture_pushcollector,
):
    fixture_iib_client.return_value.add_bundles.side_effect = (
        lambda *args, **kwargs: IIBBuildDetailsModel.from_dict(
            fake_tm.setup_task(
                *args,
                **dict(
                    list(kwargs.items()) + [("state_seq", ("in_progress", "failed"))]
                )
            )
        )
    )

    with setup_entry_point_cli(
        ("pubtools_iib", "console_scripts", "pubtools-iib-add-bundles"),
        "pubtools-iib-add-bundle",
        fixture_common_iib_op_args + ["--bundle", "bundle1"],
        {
            "OVERWRITE_FROM_INDEX_TOKEN": "overwrite_from_index_token",
        },
    ) as entry_func:
        try:
            entry_func()
            assert False, "Should have raised SystemError"
        except SystemExit:
            pass

    assert fixture_pushcollector.items == [
        operator_1_push_item_pending,
        operator_1_push_item_notpushed,
    ]

    mock_print_error_message.assert_called_once_with(
        "https://iib-server/api/v1/builds/task-4"
    )


def test_add_bundles_py(
    caplog,
    capsys,
    fixture_iib_client,
    fixture_iib_krb_auth,
    fixture_common_iib_op_args,
):
    caplog.set_level(logging.INFO)
    with setup_entry_point_py(
        ("pubtools_iib", "console_scripts", "pubtools-iib-add-bundles"),
        {
            "OVERWRITE_FROM_INDEX_TOKEN": "overwrite_from_index_token",
        },
    ) as entry_func:
        retval = entry_func(
            ["cmd"]
            + fixture_common_iib_op_args
            + ["--bundle", "bundle1", "--check-related-images", "--arch", "arch"]
        )

    assert isinstance(retval, IIBBuildDetailsModel)

    fixture_iib_client.assert_called_once_with(
        "iib-server", auth=fixture_iib_krb_auth.return_value, ssl_verify=False
    )
    fixture_iib_client.return_value.add_bundles.assert_called_once_with(
        "index-image",
        bundles=["bundle1"],
        arches=["arch"],
        binary_image="binary-image",
        overwrite_from_index=True,
        overwrite_from_index_token="overwrite_from_index_token",
        build_tags=["extra-tag-1", "extra-tag-2"],
        check_related_images=True,
    )

    task_id = retval.id
    url_msg = "IIB details: https://{}/api/v1/builds/{}".format(
        FIXTURE_IIB_SERVER, task_id
    )
    assert url_msg in [r.getMessage() for r in caplog.records]

    # neither build details nor anything else dumped into stdout
    captured = capsys.readouterr()
    assert not captured.out


def test_add_bundles_py_multiple_bundles(
    fixture_iib_client,
    fixture_iib_krb_auth,
    fixture_common_iib_op_args,
):
    with setup_entry_point_py(
        ("pubtools_iib", "console_scripts", "pubtools-iib-add-bundles"),
        {
            "OVERWRITE_FROM_INDEX_TOKEN": "overwrite_from_index_token",
        },
    ) as entry_func:
        retval = entry_func(
            ["cmd"]
            + fixture_common_iib_op_args
            + ["--bundle", "bundle1", "--bundle", "bundle2", "--arch", "arch"]
        )

    assert isinstance(retval, IIBBuildDetailsModel)

    fixture_iib_client.assert_called_once_with(
        "iib-server", auth=fixture_iib_krb_auth.return_value, ssl_verify=False
    )
    fixture_iib_client.return_value.add_bundles.assert_called_once_with(
        "index-image",
        bundles=["bundle1", "bundle2"],
        arches=["arch"],
        binary_image="binary-image",
        overwrite_from_index=True,
        overwrite_from_index_token="overwrite_from_index_token",
        build_tags=["extra-tag-1", "extra-tag-2"],
    )


@pytest.mark.parametrize(
    "extra_args,push_items,mock_calls_tester",
    [
        (
            ["--arch", "arch"],
            [operator_1_push_item_delete_pending, operator_1_push_item_deleted],
            remove_operators_mock_calls_tester,
        ),
    ],
)
def test_remove_operators_cli(
    fixture_iib_client,
    fixture_iib_krb_auth,
    fixture_common_iib_op_args,
    fixture_pushcollector,
    extra_args,
    push_items,
    mock_calls_tester,
):
    fixture_iib_client.return_value.remove_operators.side_effect = (
        lambda *args, **kwargs: IIBBuildDetailsModel.from_dict(
            fake_tm.setup_task(
                *args, **dict(list(kwargs.items()) + [("op_type", "rm")])
            )
        )
    )
    with setup_entry_point_cli(
        ("pubtools_iib", "console_scripts", "pubtools-iib-remove-operators"),
        "pubtools-iib-remove-operators",
        fixture_common_iib_op_args + ["--operator", "operator-1"] + extra_args,
        {
            "OVERWRITE_FROM_INDEX_TOKEN": "overwrite_from_index_token",
        },
    ) as entry_func:
        entry_func()
    assert fixture_pushcollector.items == push_items
    mock_calls_tester(
        fixture_iib_client,
        fixture_iib_krb_auth,
    )


@mock.patch("pubtools.iib.iib_ops.print_error_message")
def test_remove_operators_cli_error(
    mock_print_error_message,
    fixture_iib_client,
    fixture_iib_krb_auth,
    fixture_common_iib_op_args,
    fixture_pushcollector,
):
    fixture_iib_client.return_value.remove_operators.side_effect = (
        lambda *args, **kwargs: IIBBuildDetailsModel.from_dict(
            fake_tm.setup_task(
                *args,
                **dict(
                    list(kwargs.items())
                    + [("state_seq", ("in_progress", "failed")), ("op_type", "rm")]
                )
            )
        )
    )
    with setup_entry_point_cli(
        ("pubtools_iib", "console_scripts", "pubtools-iib-remove-operators"),
        "pubtools-iib-remove-operators",
        fixture_common_iib_op_args + ["--operator", "operator-1"],
        {
            "OVERWRITE_FROM_INDEX_TOKEN": "overwrite_from_index_token",
        },
    ) as entry_func:
        try:
            entry_func()
            assert False, "Should have raised SystemError"
        except SystemExit:
            pass

    assert fixture_pushcollector.items == [
        operator_1_push_item_delete_pending,
        operator_1_push_item_delete_notpushed,
    ]

    mock_print_error_message.assert_called_once_with(
        "https://iib-server/api/v1/builds/task-8"
    )


def test_remove_operators_py(
    fixture_iib_client,
    fixture_iib_krb_auth,
    fixture_common_iib_op_args,
):
    with setup_entry_point_py(
        ("pubtools_iib", "console_scripts", "pubtools-iib-remove-operators"),
        {
            "OVERWRITE_FROM_INDEX_TOKEN": "overwrite_from_index_token",
        },
    ) as entry_func:
        retval = entry_func(["cmd"] + fixture_common_iib_op_args + ["--operator", "operator-1", "--arch", "arch"])

    assert isinstance(retval, IIBBuildDetailsModel)

    remove_operators_mock_calls_tester(
        fixture_iib_client,
        fixture_iib_krb_auth,
    )


def test_add_deprecations_cli(
    fixture_iib_client,
    fixture_iib_krb_auth,
    fixture_common_iib_op_args,
    fixture_pushcollector,
):
    fixture_iib_client.return_value.add_deprecations.side_effect = (
        lambda *args, **kwargs: IIBBuildDetailsModel.from_dict(
            fake_tm.setup_task(
                *args, **dict(list(kwargs.items()) + [("op_type", "add-deprecations")])
            )
        )
    )
    with setup_entry_point_cli(
        ("pubtools_iib", "console_scripts", "pubtools-iib-add-deprecations"),
        "pubtools-iib-add-deprecations",
        fixture_common_iib_op_args + ["--operator-package", "operator-1", "--deprecation-schema", '{"a": "b"}'],
        {
            "OVERWRITE_FROM_INDEX_TOKEN": "overwrite_from_index_token",
        },
    ) as entry_func:
        entry_func()
    assert fixture_pushcollector.items == [operator_1_push_item_ad_pending, operator_1_push_item_ad]
    add_deprecations_mock_calls_tester(
        fixture_iib_client,
        fixture_iib_krb_auth,
    )


@mock.patch("pubtools.iib.iib_ops.print_error_message")
def test_add_deprecations_cli_error(
    mock_print_error_message,
    fixture_iib_client,
    fixture_iib_krb_auth,
    fixture_common_iib_op_args,
    fixture_pushcollector,
):
    fixture_iib_client.return_value.add_deprecations.side_effect = (
        lambda *args, **kwargs: IIBBuildDetailsModel.from_dict(
            fake_tm.setup_task(
                *args,
                **dict(
                    list(kwargs.items())
                    + [("state_seq", ("in_progress", "failed")), ("op_type", "add-deprecations")]
                )
            )
        )
    )
    with setup_entry_point_cli(
        ("pubtools_iib", "console_scripts", "pubtools-iib-add-deprecations"),
        "pubtools-iib-add-deprecations",
        fixture_common_iib_op_args + ["--operator-package", "operator-1", "--deprecation-schema", '{"a": "b"}'],
        {
            "OVERWRITE_FROM_INDEX_TOKEN": "overwrite_from_index_token",
        },
    ) as entry_func:
        try:
            entry_func()
            assert False, "Should have raised SystemError"
        except SystemExit:
            pass

    assert fixture_pushcollector.items == [
        operator_1_push_item_ad_pending,
        operator_1_push_item_ad_notpushed,
    ]

    mock_print_error_message.assert_called_once_with(
        "https://iib-server/api/v1/builds/task-11"
    )


def test_add_deprecations_py(
    fixture_iib_client,
    fixture_iib_krb_auth,
    fixture_common_iib_op_args,
):
    with setup_entry_point_py(
        ("pubtools_iib", "console_scripts", "pubtools-iib-add-deprecations"),
        {
            "OVERWRITE_FROM_INDEX_TOKEN": "overwrite_from_index_token",
        },
    ) as entry_func:
        retval = entry_func(["cmd"] + fixture_common_iib_op_args + ["--operator-package", "operator-1", "--deprecation-schema", '{"a": "b"}'])

    assert isinstance(retval, IIBBuildDetailsModel)

    add_deprecations_mock_calls_tester(
        fixture_iib_client,
        fixture_iib_krb_auth,
    )


def test_invalid_op(fixture_common_iib_op_args):
    try:
        _iib_op_main((), "invalid-op")
        assert False, "Should have raised"
    except ValueError:
        pass


def test_print_error_message(caplog):
    caplog.set_level(logging.INFO)
    with requests_mock.Mocker() as m:
        m.register_uri(
            "GET",
            "https://iib-test.com/api/v1/builds/5",
            json={"state_reason": "Generic IIB error"},
        )

        print_error_message("https://iib-test.com/api/v1/builds/5")

        assert (
            caplog.records[0].message
            == "IIB Failed with the error: 'Generic IIB error'"
        )
        assert caplog.records[1].message == (
            "Please check the full logs at https://iib-test.com/api/v1/builds/5/logs"
        )

        assert len(m.request_history) == 1
        assert m.request_history[0].url == "https://iib-test.com/api/v1/builds/5"



import os
import logging
import sys
from typing import Any
from argparse import Namespace, ArgumentParser

import requests
from iiblib.iib_build_details_model import IIBBuildDetailsModel

from .utils import (
    setup_iib_client,
    setup_arg_parser,
    setup_entry_point_cli,
)

import pushcollector

LOG = logging.getLogger("pubtools.iib")


CMD_ARGS = {
    ("--iib-insecure",): {
        "group": "IIB service",
        "help": "Allow unverified HTTPS connection to IIB",
        "required": False,
        "type": bool,
    },
    ("--iib-server",): {
        "group": "IIB service",
        "help": "IIB service hostname",
        "required": True,
        "type": str,
    },
    ("--iib-krb-principal",): {
        "group": "IIB service",
        "help": "IIB kerberos principal in form: name@REALM",
        "required": True,
        "type": str,
    },
    ("--iib-krb-ktfile",): {
        "group": "IIB service",
        "help": "IIB kerberos client keytab",
        "required": False,
        "type": str,
    },
    ("--index-image",): {
        "group": "IIB service",
        "help": "<hostname>/<namespace>/<image>:<tag> of index image to rebuild",
        "required": False,
        "type": str,
    },
    ("--binary-image",): {
        "group": "IIB service",
        "help": "<hostname>/<namespace>/<image>:<tag> of binary image",
        "required": False,
        "type": str,
    },
    ("--overwrite-from-index",): {
        "group": "IIB service",
        "help": (
            "overwrite from_index_image as output. If this is true,"
            " overwrite-from-index-token should also be specified."
        ),
        "required": False,
        "type": bool,
    },
    ("--overwrite-from-index-token",): {
        "group": "IIB service",
        "help": (
            "destination repo token to overwrite from_index_image"
            "If this is specified, overwrite-from-index must be set to True."
            "Or set the OVERWRITE_FROM_INDEX_TOKEN environment variable."
        ),
        "required": False,
        "type": str,
        "env_variable": "OVERWRITE_FROM_INDEX_TOKEN",
    },
    ("--build-tag",): {
        "group": "IIB service",
        "help": "extra tags to apply on built index image in temp namespace",
        "required": False,
        "type": str,
        "action": "append",
    },
    ("--build-timeout",): {
        "group": "IIB service",
        "help": "How long to wait for an IIB build before raising timeout error",
        "required": False,
        "type": str,
    },
}

ADD_CMD_ARGS = CMD_ARGS.copy()
ADD_CMD_ARGS[("--bundle",)] = {
    "group": "IIB service",
    "help": "<hostname>/<namespace>/<image>:<tag> of bundle",
    "required": False,
    "type": str,
    "action": "append",
}
ADD_CMD_ARGS[("--deprecation-list",)] = {
    "group": "IIB service",
    "help": "Comma separated list of deprecated bundles",
    "required": False,
    "type": str,
}
ADD_CMD_ARGS[("--check-related-images",)] = {
    "group": "IIB service",
    "help": "Flag to indicate if related and depending images of a bundle should be inspected",
    "required": False,
    "type": bool,
}
ADD_CMD_ARGS[("--arch",)] = {
    "group": "IIB service",
    "help": "architecture to rebuild",
    "required": False,
    "type": str,
    "action": "append",
}

RM_CMD_ARGS = CMD_ARGS.copy()
RM_CMD_ARGS[("--operator",)] = {
    "group": "IIB service",
    "help": "operator name",
    "required": True,
    "type": str,
    "action": "append",
}
RM_CMD_ARGS[("--arch",)] = {
    "group": "IIB service",
    "help": "architecture to rebuild",
    "required": False,
    "type": str,
    "action": "append",
}

ADD_DEPRECATIONS_CMD_ARGS = CMD_ARGS.copy()
ADD_DEPRECATIONS_CMD_ARGS[("--deprecation-schema",)] = {
    "group": "IIB service",
    "help": "JSON formatted deprecation schema",
    "required": True,
    "type": str,
}
ADD_DEPRECATIONS_CMD_ARGS[("--operator-package",)] = {
    "group": "IIB service",
    "help": "operator name",
    "required": True,
    "type": str,
}


def push_items_from_build(
    build_details: IIBBuildDetailsModel, state: str
) -> list[dict[Any, Any]]:
    ret = []
    if build_details.request_type == "add":
        for operator, bundles in build_details.bundle_mapping.items():
            for bundle in bundles:
                item = {
                    "state": state,
                    "origin": build_details.from_index or "scratch",
                    "src": bundle,
                    "filename": operator,
                    "dest": "redhat-operator-index",
                    "build": build_details.index_image,
                    "signing_key": None,
                    "checksums": None,
                }
                ret.append(item)
    elif build_details.request_type == "rm":
        for operator in build_details.removed_operators:
            item = {
                "state": state,
                "origin": build_details.from_index,
                "src": None,
                "filename": operator,
                "dest": "redhat-operator-index",
                "build": build_details.index_image,
                "signing_key": None,
                "checksums": None,
            }
            ret.append(item)
    elif build_details.request_type == "add-deprecations":
        item = {
            "state": state,
            "origin": build_details.from_index,
            "src": None,
            "filename": build_details.operator_package,
            "dest": "redhat-operator-index",
            "build": build_details.index_image,
            "signing_key": None,
            "checksums": None,
        }
        ret.append(item)

    return ret


def process_parsed_args(parsed_args: Namespace, args: dict[Any, Any]) -> Namespace:
    for aliases, arg_data in args.items():
        named_alias = [
            x.lstrip("-").replace("-", "_") for x in aliases if x.startswith("--")
        ][0]
        if arg_data.get("env_variable"):
            if not getattr(parsed_args, named_alias) and os.environ.get(
                arg_data["env_variable"]
            ):
                setattr(
                    parsed_args,
                    named_alias,
                    os.environ.get(arg_data["env_variable"]),
                )
    return parsed_args


def _iib_op_main(
    args: Namespace,
    operation: str | None = None,
    items_final_state: str = "PUSHED",
) -> list[dict[Any, Any]] | Any:
    if operation not in ("add_bundles", "remove_operators", "add_deprecations"):
        raise ValueError("Must set iib operation")

    pc = pushcollector.Collector.get()
    LOG.debug("Initializing iib client")
    iib_c = setup_iib_client(args)
    LOG.debug("Request to rebuild %s", args.index_image)

    bundle_op = getattr(iib_c, operation)

    extra_args = {}
    if operation == "add_bundles":
        if args.deprecation_list:
            extra_args["deprecation_list"] = args.deprecation_list.split(",")
        if args.check_related_images:
            extra_args["check_related_images"] = args.check_related_images
        extra_args["arches"] = args.arch
        extra_args["bundles"] = args.bundle
    if operation == "remove_operators":
        extra_args["arches"] = args.arch
        extra_args["operators"] = args.operator
    if operation == "add_deprecations":
        extra_args["operator_package"] = args.operator_package
        extra_args["deprecation_schema"] = args.deprecation_schema

    if args.binary_image:
        extra_args["binary_image"] = args.binary_image

    if args.overwrite_from_index:
        extra_args["overwrite_from_index"] = args.overwrite_from_index

    if args.overwrite_from_index_token:
        extra_args["overwrite_from_index_token"] = args.overwrite_from_index_token

    if args.build_tag:
        extra_args["build_tags"] = args.build_tag

    build_details = bundle_op(args.index_image, **extra_args)

    push_items = push_items_from_build(build_details, "PENDING")
    LOG.debug("Updating push items")
    pc.update_push_items(push_items)

    build_details = iib_c.wait_for_build(build_details)

    build_details_url = _make_iib_build_details_url(args.iib_server, build_details.id)
    LOG.info("IIB details: %s", build_details_url)

    if build_details.state == "failed":
        LOG.error("IIB operation failed")
        print_error_message(build_details_url)
        push_items = push_items_from_build(build_details, "NOTPUSHED")
        pc.update_push_items(push_items)
        sys.exit(1)

    push_items = push_items_from_build(build_details, items_final_state)
    pc.update_push_items(push_items)
    LOG.info("IIB build finished")
    return build_details


def make_add_bundles_parser() -> ArgumentParser:
    return setup_arg_parser(ADD_CMD_ARGS)


def make_rm_operators_parser() -> ArgumentParser:
    return setup_arg_parser(RM_CMD_ARGS)


def make_add_deprecations_parser() -> ArgumentParser:
    return setup_arg_parser(ADD_DEPRECATIONS_CMD_ARGS)


def add_bundles_main(sysargs: list[str] | None = None) -> list[dict[Any, Any]]:
    logging.basicConfig(level=logging.INFO)

    parser = make_add_bundles_parser()
    if sysargs:
        args = parser.parse_args(sysargs[1:])
    else:
        args = parser.parse_args()
    process_parsed_args(args, ADD_CMD_ARGS)

    return _iib_op_main(args, "add_bundles")


def remove_operators_main(
    sysargs: list[str] | None = None,
) -> list[dict[Any, Any]]:
    logging.basicConfig(level=logging.INFO)

    parser = make_rm_operators_parser()
    if sysargs:
        args = parser.parse_args(sysargs[1:])
    else:
        args = parser.parse_args()
    process_parsed_args(args, RM_CMD_ARGS)

    return _iib_op_main(args, "remove_operators", "DELETED")


def add_deprecations_main(
    sysargs: list[str] | None = None,
) -> list[dict[Any, Any]]:
    logging.basicConfig(level=logging.INFO)

    parser = make_add_deprecations_parser()
    if sysargs:
        args = parser.parse_args(sysargs[1:])
    else:
        args = parser.parse_args()
    process_parsed_args(args, ADD_DEPRECATIONS_CMD_ARGS)

    return _iib_op_main(args, "add_deprecations")


def _make_iib_build_details_url(host: str, task_id: str) -> str:
    return "https://%s/api/v1/builds/%s" % (host, task_id)


def print_error_message(iib_build_url: str) -> None:
    """
    Construct and print an error message for IIB failure.

    Args:
        iib_build_url (str): URL of the IIB build.
    """
    res = requests.get(iib_build_url, timeout=30).json()

    LOG.error("IIB Failed with the error: '%s'", res["state_reason"])
    LOG.error("Please check the full logs at %s", f"{iib_build_url.rstrip('/')}/logs")
